unit Resolver;

{*****************************************************************************
 *                                                                           *
 * Diese Unit enthlt die Delphi-VCL-Komponente TResolver.                   *
 * Sie basiert auf den Socketkomponenten aus der Unit ScktComp von Delphi 3  *
 * und realisiert eine DNS-Abfrage bei minimalem Funktionsumfang.            *
 *                                                                           *
 *(Details zum DNS-Protokoll siehe auch www.hummel1.de/prog/tcpip/dnsrfc.htm)*
 *                                                                           *
 *****************************************************************************}

interface

uses Classes, ScktComp, ExtCtrls;

type TResolverErrorEvent = procedure(Sender:TObject; ErrorMsg:String) of object;

type TResolver = class(TComponent)
 private
  FActive: Boolean;                                       // Verbindungszustand
  FAnswer: String;                                          // Rckgabeergebnis
  FID: Word;                                                 // Query-ID-Nummer
  FInLength: Integer;                                // empfangene Antwortlnge
  FInString: String;                                         // Empfangsantwort
  FQueryLength: Integer;                                        // Abfragelnge
  FQueryTyp: Integer;                                             // Abfragetyp
  FServer: String;                                          //remote DNS-Server
  FSock: TClientSocket;                                         // ClientSocket
  FTimer: TTimer;                                              // TimeOut-Timer
  FOnConnect: TNotifyEvent;
  FOnAnswer: TNotifyEvent;
  FOnError: TResolverErrorEvent;
  FOnTimeOut: TNotifyEvent;
  function  GetTimeOut: Cardinal;
  procedure SetTimeOut(Time: Cardinal);
  procedure SockConnect(Sender: TObject; Socket: TCustomWinSocket);
  procedure SockRead(Sender: TObject; Socket: TCustomWinSocket);
  procedure SockError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
  procedure TimerIsOut(Sender: TObject);
 public
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  property Active: Boolean read FActive;                  // Verbindungszustand
  property Answer: String read FAnswer;                       // Afrageergebnis
  procedure Connect;                                // verbinden mit DNS-Server
  procedure Close;                                      // Verbindung schlieen
  procedure Query(DNS: String);                          // Abfrage durchfhren
 published
  property Server: String read FServer write FServer;
  property TimeOut: Cardinal read GetTimeOut write SetTimeOut default 15000;
  property OnConnect: TNotifyEvent read FOnConnect write FOnConnect;
  property OnError: TResolverErrorEvent read FOnError write FOnError;
  property OnAnswer: TNotifyEvent read FOnAnswer write FOnAnswer;
  property OnTimeOut: TNotifyEvent read FOnTimeOut write FOnTimeOut;
end;

procedure Register;

implementation

uses SysUtils;

procedure Register;
begin
  RegisterComponents('Internet', [TResolver]);
end;

constructor TResolver.Create(AOwner: TComponent);
begin
 inherited Create(AOwner);
 FActive:=false;
 FAnswer:='';
 FInString:='';
 FServer:='';
 FSock:=TClientSocket.Create(Self);                    // ClientSocket erzeugen
 FSock.ClientType:=ctNonBlocking;
 FSock.Port:=53;                                                  // DNS-Dienst
 FSock.OnConnect:=SockConnect;
 FSock.OnRead:=SockRead;
 FSock.OnError:=SockError;
 FSock.Socket.OnErrorEvent:=SockError;
 FTimer:=TTimer.Create(Self);                         // TimeOut-Timer erzeugen
 FTimer.Enabled:=false;
 FTimer.Interval:=15000;
 FTimer.OnTimer:=TimerIsOut;
 FQueryLength:=999;                                   // Dummy gegen Lesefehler
 Randomize;                                    // Zufallsgenerator fr Query-ID
end;

destructor TResolver.Destroy;
begin
 FTimer.Enabled:=false;
 if Active then FSock.Close;
 FTimer.Free;
 FSock.Free;
 inherited Destroy;
end;

function TResolver.GetTimeOut: Cardinal;                  // TimeOut-Zeit lesen
begin Result:=FTimer.Interval end;

procedure TResolver.SetTimeOut(Time: Cardinal);          // TimeOut-Zeit ndern
begin FTimer.Interval:=Time end;

procedure TResolver.Connect;                           // Verbindung herstellen
var nDummy: Integer; sOldSvr: String;
begin
 if Copy(Server,1,1)>='A' then sOldSvr:=FSock.Host else sOldSvr:=FSock.Address;
 if Active and (Server<>sOldSvr) then begin FSock.Close; FActive:=false end;
 if not Active then begin                               // nur einmal verbinden
  if Copy(Server,1,1)>='A' then FSock.Host:=Server else FSock.Address:=Server;
  try
   FSock.Open;
   FActive:=true;                           // nur wenn FSock.Open geklappt hat
  except
   on E:ESocketError do SockError(Self, FSock.Socket, eeConnect, nDummy);
  end; // try
 end; // if not(Active)
end;

procedure TResolver.Close;                                // Verbindung beenden
begin
 if FTimer.Enabled then FTimer.Enabled:=false;
 if FSock.Active then FSock.Close;
 FActive:=false;
end;

procedure TResolver.Query(DNS: String);                   // Abfrage generieren
var sSend, sDns, sBuf, sBu2: String; p: Integer;
begin
 sDns:=DNS; sBuf:=''; sBu2:='';
 if (not FActive)or(Length(sDns)=0) then begin             // Abfrage unmglich
  if Assigned(FOnError) then FOnError(Self, 'Abfragefehler');  // Fehlermeldung
  if Assigned(FOnAnswer) then begin                             // Null-Antwort
   FAnswer:='';
   FOnAnswer(Self);
  end;
  Exit;
 end;
 FID:=Random(62000);                           //Query-ID generieren und merken
 sSend:=Chr(FID div 256)+Chr(FID mod 256);
 sSend:=sSend+Chr(01)+Chr(00);                           // Query mit Rekursion
 sSend:=sSend+Chr(00)+Chr(01);                                    // eine Query
 sSend:=sSend+Chr(00)+Chr(00)+Chr(00)+Chr(00)+Chr(00)+Chr(00); // Header fertig
 if Copy(sDns,1,1)>='A' then begin                      // Anfrage Host-Address
  FQueryTyp:=1;                                             // Type HostAddress
  while Length(sDns)>0 do begin
   p:=Pos('.',sDns);
   if p=0 then p:=Length(sDns)+1;
   sBuf:=Copy(sDns,1,p-1);
   Delete(sDns,1,p);
   sSend:=sSend+Chr(Length(sBuf))+sBuf;           // Host-String zusammensetzen
  end;
 end else begin                                  // Anfrage Domain-Name-Pointer
  FQueryTyp:=12;                                            // Type Pointer PTR
  while Length(sDns)>0 do begin
   p:=Pos('.',sDns);
   if p=0 then p:=Length(sDns)+1;
   sBuf:=Copy(sDns,1,p-1);
   Delete(sDns,1,p);
   sBu2:=Chr(Length(sBuf))+sBuf + sBu2;             // IP-String zusammensetzen
  end;
  sSend:=sSend+sBu2+Chr(07)+'in-addr'+Chr(04)+'arpa';  // anhngen IN-ADDR.ARPA
 end;
 sSend:=sSend+Chr(00);                                                  // Root
 sSend:=sSend+Chr(00)+Chr(FQueryTyp);                                  // QType
 sSend:=sSend+Chr(00)+Chr(01);                                // Class InterNet
 FQueryLength:=Length(sSend);
 sSend:=Chr(FQueryLength div 256)+Chr(FQueryLength mod 256)+sSend; // IP-Header
 Inc(FQueryLength,2);                                    // Korrektur der Lnge
 FInString:='';
 FInLength:=0;
 FTimer.Enabled:=true;                                 // Timeout-Timer starten
 FSock.Socket.SendText(sSend);                                // Anfrage senden
end;

procedure TResolver.SockRead(Sender: TObject; Socket: TCustomWinSocket);
var i, j, nRDLen: Integer; sBuf: String;                   // bei Socketempfang
begin
 sBuf:=Socket.ReceiveText;
 if FTimer.Enabled then FInString:=FInString+sBuf;   // nach TimeOut nicht mehr
 if Assigned(FOnAnswer) and (Length(FInString)>=FQueryLength) then begin
  if FInLength=0 then FInLength:=Ord(FInString[1])shl 8+Ord(FInString[2])+2;
  if Length(FInString)=FInLength then begin         // Antwort komplett gelesen
   if FID<>(Ord(FInString[3])shl 8+Ord(FInString[4])) then begin  // falsche ID
    FInString:='';
    FInLength:=0;
    Exit;
   end;
   FTimer.Enabled:=false;                               // TimeOutTimer stoppen
   sBuf:='';
   if (Ord(FInString[6])mod 16)=0 then begin                  // ResponseCode=0
    Delete(FInString,1,FQueryLength);                      // Query abschneiden
    if FQueryTyp=1 then                                 // es war Address-Query
     repeat
      nRDLen:=0;                                // Dummy, wegen Compilerwarnung
      i:=Pos(Chr(00)+Chr(01)+Chr(00)+Chr(01),FInString); // Type1+Class1 suchen
      if i>0 then begin
       nRDLen:=Ord(FInString[i+8])shl 8+Ord(FInString[i+9]);        // RDLength
       if nRDLen=4 then begin
        for j:=i+10 to i+13 do sBuf:=sBuf+IntToStr(Ord(FInString[j]))+'.';
        Delete(sBuf,Length(sBuf),1);                // letzten Punkt wieder weg
       end else Delete(FInString,1,i+1)       // falsche Fundstelle, neu suchen
      end;
     until (nRDLen=4)or(i=0)
    else begin                                          // es war Pointer-Query
     i:=Pos(Chr(00)+Chr(12)+Chr(00)+Chr(01),FInString); // Type12+Class1 suchen
     if i>0 then begin
      nRDLen:=Ord(FInString[i+8])shl 8+Ord(FInString[i+9]);         // RDLength
      FInString:=Copy(FInString,i+10,nRDLen);       // RDLength herausschneiden
      while Length(FInString)>1 do begin
       i:=Ord(FInString[1]);                       // Lnge des nchsten Labels
       sBuf:=sBuf+Copy(FInString,2,i);                        // Label anhngen
       Delete(FInString,1,i+1);               // Label in InString wegschneiden
       if Length(FInString)>1 then sBuf:=sBuf+'.';            // Punkt anhngen
      end;
     end;
    end; // if QueryType 1 or 12
   end; //if ResponseCode=0
   FAnswer:=sBuf;
   FOnAnswer(Self);
  end; // if Length=FInLength
 end; // if Assigned
end;

procedure TResolver.TimerIsOut(Sender: TObject);  // Zeitberschreitung Abfrage
begin
 FTimer.Enabled:=false;
 if Assigned(FOnTimeOut) then FOnTimeOut(Self);
 FInString:='';
 FAnswer:='';
 if Assigned(FOnAnswer) then FOnAnswer(Self);   // leeres Ergebnis zurckmelden
end;

procedure TResolver.SockConnect(Sender: TObject; Socket: TCustomWinSocket);
begin if Assigned(FOnConnect) then FOnConnect(Self) end;  // bei Connect-Erfolg

procedure TResolver.SockError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
var sMsg: String;
begin
 FActive:=False;                                         // Status zurcksetzen
 if Assigned(FOnError) then begin
  case ErrorEvent of
   eeSend   : sMsg:='Sendefehler';
   eeReceive: sMsg:='Empfangsfehler';
   eeConnect: sMsg:='Verbindungsfehler';
   eeDisconnect: sMsg:='Abmeldefehler';
   eeAccept : sMsg:='Annahmefehler';
   else sMsg:='allgemeiner Fehler';
  end;
  FOnError(Self, sMsg);                                        // Fehler melden
 end;
 ErrorCode:=0;                                   // kein Error, keine Exception
 FSock.Close;                                      // Verbindung lieber trennen
end;

end.
